import numpy as np
import pandas as pd
import networkx as nx
from networkx.algorithms.approximation.traveling_salesman import christofides


def floyd_algorithm(cost_matrix):

    n = len(cost_matrix)
    shortest_path_matrix = [[0] * n for _ in range(n)]

    for i in range(n):
        for j in range(n):
            shortest_path_matrix[i][j] = cost_matrix[i][j]

    for k in range(n):
        for i in range(n):
            for j in range(n):
                if i!=j:
                    shortest_path_matrix[i][j] = min(shortest_path_matrix[i][j], shortest_path_matrix[i][k] + shortest_path_matrix[k][j])
                else:
                    shortest_path_matrix[i][j] = np.inf

    return shortest_path_matrix


def solve_christofides(distances):

    cost_matrix=distances
    cost_matrix=floyd_algorithm(cost_matrix)

    # Create an empty graph
    graph = nx.Graph()

    # Add nodes to the graph
    num_nodes = len(cost_matrix)
    graph.add_nodes_from(range(num_nodes))

    for i in range(num_nodes):
        for j in range(i + 1, num_nodes):
            distance = cost_matrix[i][j]
            graph.add_edge(i, j, weight=distance)
    tour=christofides(graph)
    costs=0
    for i in range(len(cost_matrix)):
        costs+=cost_matrix[tour[i]][tour[i+1]]

    for i in range(len(tour)):
        tour[i]+=1
    return tour, costs
