import Christofides
import random
import pandas as pd
import numpy as np
import time
def floyd_algorithm(cost_matrix):

    n = len(cost_matrix)
    shortest_path_matrix = [[0] * n for _ in range(n)]

    for i in range(n):
        for j in range(n):
            shortest_path_matrix[i][j] = cost_matrix[i][j]

    for k in range(n):
        for i in range(n):
            for j in range(n):
                if i!=j:
                    shortest_path_matrix[i][j] = min(shortest_path_matrix[i][j], shortest_path_matrix[i][k] + shortest_path_matrix[k][j])
                else:
                    shortest_path_matrix[i][j] = np.inf

    return shortest_path_matrix



def initialize_solution(distances):
    initial_tour_data=Christofides.solve_christofides(distances)
    initial_tour=initial_tour_data[0]
    initial_costs=initial_tour_data[1]
    for i in range(len(initial_tour)):
        initial_tour[i]+=-1
    return initial_tour, initial_costs

def calculate_cost(solution, distances):
    # Berechne die Kosten der Lösung
    # Gib die Kosten als Wert zurück
    costs=0
    for i in range(len(solution)-1):
        costs+=distances[solution[i]][solution[i+1]]
    return costs
    

def termination_criteria_not_met(iteration_counter, max_counter, max_time, current_time):
    # Überprüfe die Abbruchbedingung des Algorithmus
    # Gib True zurück, wenn der Algorithmus weiter ausgeführt werden soll, sonst False
    if iteration_counter > max_counter or current_time > max_time:
        return False
    else:
        return True

def destroy(solution, d):
    # Führe die Destruction-Phase aus
    # Zerstöre Teile der aktuellen Lösung
    # Gib die zerstörte Lösung zurück
    destroyed_sol=solution
    removed=list()
    for i in range(d):
        idx=random.randint(0,len(destroyed_sol)-1)
        if idx==0 or idx==len(destroyed_sol)-1:
            item=destroyed_sol.pop(0)
            item=destroyed_sol.pop(len(destroyed_sol)-1)
            destroyed_sol.append(destroyed_sol[0])
            removed.append(item)
        else:
            item=destroyed_sol.pop(idx)
            removed.append(item)
    return destroyed_sol, removed


def reconstruct(destroyed_solution, removed, d, distances):
    # Führe die Reconstruction-Phase aus
    # Rekonstruiere die zerstörte Lösung
    # Gib die rekonstruierte Lösung zurück
    for i in range(d):
        working_sol=destroyed_solution
        cost_dict=dict()
        item=removed.pop(0)
        for j in range(len(working_sol)-1):
            if j==0:
                
                working_sol.insert(j, item)
                working_sol.pop()
                working_sol.insert(len(working_sol), item)
                
            else:
                working_sol.insert(j, item)
            
            costs=calculate_cost(working_sol, distances)
            cost_dict[j]=costs
            if j==0:
                working_sol.pop()
                working_sol.pop(0)
                working_sol.append(working_sol[0])
            else:
                working_sol.pop(j)

        
        min_idx=min(cost_dict, key=cost_dict.get)
        
        min_cost=min(cost_dict.values())
        

        if min_idx==0 or min_idx==len(working_sol)-1:
            destroyed_solution.insert(0, item)
            destroyed_solution.pop()
            destroyed_solution.append(item)
        else:
            destroyed_solution.insert(min_idx, item)
        
    
    des_sol_cost=min_cost
    return destroyed_solution, des_sol_cost 

def accept(reconstructed_solution, reconstructed_cost, current_solution, current_best_costs, iteration_counter):
    # Überprüfe, ob die rekonstruierte Lösung akzeptiert werden soll
    # Vergleiche sie mit der aktuellen Lösung
    # Gib True zurück, wenn die rekonstruierte Lösung akzeptiert werden soll, sonst False
    if reconstructed_cost < current_best_costs:
        current_solution=reconstructed_solution
        current_best_costs=reconstructed_cost
        iteration_counter=0
        return current_solution, current_best_costs, iteration_counter
    else:
        iteration_counter+=1
        return current_solution, current_best_costs, iteration_counter

def greedy_algorithm(distances, max_time, max_counter, d):
    # Initialisierung
    solution, cost = initialize_solution(distances)
    best_solution = solution
    best_cost = cost
    

    
    iteration_counter=0
    start_time=time.time()
    runtime=0
    while termination_criteria_not_met(iteration_counter, max_counter, max_time, runtime):
        # Destruction
        destroyed_solution, removed = destroy(solution, d)

        # Reconstruction
        reconstructed_solution, reconstructed_cost = reconstruct(destroyed_solution, removed, d, distances)

        # Akzeptanz
        best_solution, best_cost, iteration_counter=accept(reconstructed_solution, reconstructed_cost, best_solution, best_cost, iteration_counter)
        
        
        current_time=time.time()
        runtime=current_time-start_time
    
    for i in range(len(best_solution)):
        best_solution[i]+=1
    return best_solution, best_cost

# Beispiel für die Verwendung des Greedy-Algorithmus
distances=pd.read_csv("travel_times_capitals.csv", delimiter=";", index_col=0).values.tolist()
distances=floyd_algorithm(distances)
